/**
 *
 * CallableStatement.java
 *
 * Copyright 1998, Junus Junarto D
 *                 (junus@its.ac.id)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * @version 0.1.0 1-MARCH-1998
 *
 * A Connection represents a session with a specific database.  Within the
 * context of a Connection, SQL statements are executed and results are
 * returned.
 *
 * A Connection's database is able to provide information describing
 * its tables, its supported SQL grammar, its stored procedures, the
 * capabilities of this connection, etc.  This information is obtained
 * with the getMetaData method.
 *
 * <B>Note:</B> By default, the Connection automatically commits changes
 * after executing each statement.  If auto-commit has been disabled, an
 * explicit commit must be done or database changes will not be saved.
 *
 * @see java.sql.Connection
 */

package iddb;

import java.io.*;
import java.lang.*;
import java.net.*;
import java.util.*;
import java.sql.*;
//import iddb.*;

public class Connection implements java.sql.Connection
{
  protected IDDBStream iddb_stream;

  private String IDDB_HOST;
  private int IDDB_PORT;
  private String IDDB_USER;
  private String IDDB_PASSWORD;
  private String IDDB_DATABASE;
  private boolean IDDB_STATUS;

  public boolean CONNECTION_OK = true;
  public boolean CONNECTION_BAD = false;

  private int STARTUP_CODE = 9301;

  private boolean autoCommit = true;
  private boolean readOnly = false;

  protected Driver this_driver;
  private String this_url;
  private String cursor = null;	// The positioned update cursor name

  private IDDBLocalConnection IDDB_local_conn;
  boolean local=false;
  private int vcTransactionIsolation=Connection.TRANSACTION_SERIALIZABLE;
  byte[] extendedData;

  boolean vcCommit=false;

    String encrypt(String user, String passin)
    {
        if(passin.length()!=0)
        {
            byte[] buser=user.getBytes();
            byte[] bpassin=passin.getBytes();
            int totaluser=0;
            for(int i=0;i<buser.length;i++)
                totaluser+=buser[i];
            totaluser%=bpassin.length;
            byte[] bresult=new byte[bpassin.length];
            for(int i=0;i<bpassin.length;i++)
                bresult[(totaluser+i)%bpassin.length]=bpassin[bpassin.length-i-1];
            return new String(bresult);
        }
        return "";
    }

  /**
   * Connect to a IDDB database back end.
   *
   * @param host the hostname of the database back end
   * @param port the port number of the postmaster process
   * @param info a Properties[] thing of the user and password
   * @param database the database to connect to
   * @param u the URL of the connection
   * @param d the Driver instantation of the connection
   * @return a valid connection profile
   * @exception SQLException if a database access error occurs
   */
  public Connection(String host, int port, Properties info, String database, String url, Driver d) throws SQLException
  {
    int len = 98;			// Length of a startup packet

    this_driver = d;
    this_url = new String(url);
    IDDB_DATABASE = new String(database);
    IDDB_PASSWORD = new String(info.getProperty("password"));
    IDDB_USER = new String(info.getProperty("user"));
    IDDB_PORT = port;
    IDDB_HOST = new String(host);
    IDDB_STATUS = CONNECTION_BAD;


    try
      {
	iddb_stream = new IDDBStream(host, port);
      } catch (IOException e) {
	throw new SQLException ("Connection failed: " + e.toString());
      }
      // Now we need to construct and send a startup packet
      try
	{
	  iddb_stream.SendInteger(len, 4);			    len -= 4;
	  iddb_stream.SendInteger(STARTUP_CODE, 4);		len -= 4;
	  iddb_stream.Send(database.getBytes(), 20);	len -= 20;
	  iddb_stream.Send(IDDB_USER.getBytes(), 20);  len -= 20;
	  iddb_stream.Send(encrypt("IdDb",IDDB_PASSWORD).getBytes(), len);

	  cek_init_status();
	} catch (IOException e) {
	  throw new SQLException("Connection failed: " + e.toString());
	}
//	ExecSQL(" ");				// Test connection
	IDDB_STATUS = CONNECTION_OK;

//    System.out.println("Mode IDDB Local : "+local);
  }

    public Connection(IDDBLocalConnection il, Driver d) throws SQLException
    {
        local=true;
        IDDB_local_conn=il;
        this_driver = d;

//        System.out.println("Mode IDDB Local : "+local);
    }

    public void cek_init_status() throws IOException
    {
	    char c=(char)iddb_stream.ReceiveChar();
	    switch (c)
	    {
	        case 'O':
//	            System.out.println("Ok Connected :)");
	            break;
	        case 'E':
	            String strbuf=iddb_stream.ReceiveString(1000);
		    iddb_stream.close();
	  	    throw new IOException(strbuf);
	    }

    }
  /**
   * SQL statements without parameters are normally executed using
   * Statement objects.  If the same SQL statement is executed many
   * times, it is more efficient to use a PreparedStatement
   *
   * @return a new Statement object
   * @exception SQLException passed through from the constructor
   */
  public java.sql.Statement createStatement() throws SQLException
  {
    if(!local)
      return new Statement(this);
    else
      return new Statement(this,IDDB_local_conn.createStatement(),IDDB_USER);
  }

  /**
   * A SQL statement with or without IN parameters can be pre-compiled
   * and stored in a PreparedStatement object.  This object can then
   * be used to efficiently execute this statement multiple times.
   *
   * <B>Note:</B> This method is optimized for handling parametric
   * SQL statements that benefit from precompilation if the drivers
   * supports precompilation.  IDDB does not support precompilation.
   * In this case, the statement is not sent to the database until the
   * PreparedStatement is executed.  This has no direct effect on users;
   * however it does affect which method throws certain SQLExceptions
   *
   * @param sql a SQL statement that may contain one or more '?' IN
   *	parameter placeholders
   * @return a new PreparedStatement object containing the pre-compiled
   *	statement.
   * @exception SQLException if a database access error occurs.
   */
  public java.sql.PreparedStatement prepareStatement(String sql) throws SQLException
  {
    return new PreparedStatement(this, sql);
  }

  /**
   * A SQL stored procedure call statement is handled by creating a
   * CallableStatement for it.  The CallableStatement provides methods
   * for setting up its IN and OUT parameters and methods for executing
   * it.
   *
   * <B>Note:</B> This method is optimised for handling stored procedure
   * call statements.  Some drivers may send the call statement to the
   * database when the prepareCall is done; others may wait until the
   * CallableStatement is executed.  This has no direct effect on users;
   * however, it does affect which method throws certain SQLExceptions
   *
   * @param sql a SQL statement that may contain one or more '?' parameter
   *	placeholders.  Typically this statement is a JDBC function call
   *	escape string.
   * @return a new CallableStatement object containing the pre-compiled
   *	SQL statement
   * @exception SQLException if a database access error occurs
   */
  public java.sql.CallableStatement prepareCall(String sql) throws SQLException
  {
    throw new SQLException("Callable Statements are not supported at this time");
    //		return new CallableStatement(this, sql);
  }

  /**
   * A driver may convert the JDBC sql grammar into its system's
   * native SQL grammar prior to sending it; nativeSQL returns the
   * native form of the statement that the driver would have sent.
   *
   * @param sql a SQL statement that may contain one or more '?'
   *	parameter placeholders
   * @return the native form of this statement
   * @exception SQLException if a database access error occurs
   */
  public String nativeSQL(String sql) throws SQLException
  {
    return sql;
  }

  /**
   * If a connection is in auto-commit mode, than all its SQL
   * statements will be executed and committed as individual
   * transactions.  Otherwise, its SQL statements are grouped
   * into transactions that are terminated by either commit()
   * or rollback().  By default, new connections are in auto-
   * commit mode.  The commit occurs when the statement completes
   * or the next execute occurs, whichever comes first.  In the
   * case of statements returning a ResultSet, the statement
   * completes when the last row of the ResultSet has been retrieved
   * or the ResultSet has been closed.  In advanced cases, a single
   * statement may return multiple results as well as output parameter
   * values.  Here the commit occurs when all results and output param
   * values have been retrieved.
   *
   * @param autoCommit - true enables auto-commit; false disables it
   * @exception SQLException if a database access error occurs
   */
  public void setAutoCommit(boolean autoCommit) throws SQLException
  {
    try
    {
    if (this.autoCommit == autoCommit)
      return;
    iddb_stream.SendChar('A');
    if (autoCommit)
        iddb_stream.SendChar('Y');
    else
        iddb_stream.SendChar('N');
    this.autoCommit = autoCommit;
	    } catch (IOException e) {}
  }

  /**
   * gets the current auto-commit state
   *
   * @return Current state of the auto-commit mode
   * @exception SQLException (why?)
   * @see setAutoCommit
   */
  public boolean getAutoCommit() throws SQLException
  {
    return this.autoCommit;
  }

  /**
   * The method commit() makes all changes made since the previous
   * commit/rollback permanent and releases any database locks currently
   * held by the Connection.  This method should only be used when
   * auto-commit has been disabled.  (If autoCommit == true, then we
   * just return anyhow)
   *
   * @exception SQLException if a database access error occurs
   * @see setAutoCommit
   */
  public void commit() throws SQLException
  {
if(!vcCommit)
{
    vcCommit=true;
    if (autoCommit)
      return;
    ExecSQL("commit");
    vcCommit=false;
}
  }

  /**
   * The method rollback() drops all changes made since the previous
   * commit/rollback and releases any database locks currently held by
   * the Connection.
   *
   * @exception SQLException if a database access error occurs
   * @see commit
   */
  public void rollback() throws SQLException
  {
    if (autoCommit)
      return;
    ExecSQL("rollback");
  }

  /**
   * In some cases, it is desirable to immediately release a Connection's
   * database and JDBC resources instead of waiting for them to be
   * automatically released (cant think why off the top of my head)
   *
   * <B>Note:</B> A Connection is automatically closed when it is
   * garbage collected.  Certain fatal errors also result in a closed
   * connection.
   *
   * @exception SQLException if a database access error occurs
   */
  public void close() throws SQLException
  {
    if (iddb_stream != null)
    {
	    try
	    {
//	        if(!vcCommit)
//	        {
	            commit();
	            iddb_stream.SendChar('S');
	            int c = iddb_stream.ReceiveChar();
// System.out.println("ini close Connection "+c);
                if(c!='S')
                    throw new SQLException("Close Connection failed: ");

//	        }
	        iddb_stream.close();
	    } catch (IOException e) {}
	    iddb_stream = null;
    }
  }

  /**
   * Tests to see if a Connection is closed
   *
   * @return the status of the connection
   * @exception SQLException (why?)
   */
  public boolean isClosed() throws SQLException
  {
    return (iddb_stream == null);
  }

  /**
   * A connection's database is able to provide information describing
   * its tables, its supported SQL grammar, its stored procedures, the
   * capabilities of this connection, etc.  This information is made
   * available through a DatabaseMetaData object.
   *
   * @return a DatabaseMetaData object for this connection
   * @exception SQLException if a database access error occurs
   */
  public java.sql.DatabaseMetaData getMetaData() throws SQLException
  {
    return new DatabaseMetaData(this);
  }

  /**
   * You can put a connection in read-only mode as a hunt to enable
   * database optimizations
   *
   * <B>Note:</B> setReadOnly cannot be called while in the middle
   * of a transaction
   *
   * @param readOnly - true enables read-only mode; false disables it
   * @exception SQLException if a database access error occurs
   */
  public void setReadOnly (boolean readOnly) throws SQLException
  {
    this.readOnly = readOnly;
  }

  /**
   * Tests to see if the connection is in Read Only Mode.  Note that
   * we cannot really put the database in read only mode, but we pretend
   * we can by returning the value of the readOnly flag
   *
   * @return true if the connection is read only
   * @exception SQLException if a database access error occurs
   */
  public boolean isReadOnly() throws SQLException
  {
    return readOnly;
  }

  /**
   * A sub-space of this Connection's database may be selected by
   * setting a catalog name.  If the driver does not support catalogs,
   * it will silently ignore this request
   *
   * @exception SQLException if a database access error occurs
   */
  public void setCatalog(String catalog) throws SQLException
  {
    // No-op
  }

  /**
   * Return the connections current catalog name, or null if no
   * catalog name is set, or we dont support catalogs.
   *
   * @return the current catalog name or null
   * @exception SQLException if a database access error occurs
   */
  public String getCatalog() throws SQLException
  {
    return null;
  }

  /**
   * You can call this method to try to change the transaction
   * isolation level using one of the TRANSACTION_* values.
   *
   * <B>Note:</B> setTransactionIsolation cannot be called while
   * in the middle of a transaction
   *
   * @param level one of the TRANSACTION_* isolation values with
   *	the exception of TRANSACTION_NONE; some databases may
   *	not support other values
   * @exception SQLException if a database access error occurs
   * @see java.sql.DatabaseMetaData#supportsTransactionIsolationLevel
   */
  public void setTransactionIsolation(int level) throws SQLException
  {
    try
    {
    if((level!=Connection.TRANSACTION_READ_UNCOMMITTED)&&(level!=Connection.TRANSACTION_NONE))
        throw new SQLException("That Transaction Isolation Level is not implemented");
    vcTransactionIsolation=level;
	iddb_stream.SendChar('L');
    switch(level)
    {
      case Connection.TRANSACTION_READ_UNCOMMITTED:
	    iddb_stream.SendInteger(Connection.TRANSACTION_READ_UNCOMMITTED, 4);
	    break;
      case Connection.TRANSACTION_NONE:
	    iddb_stream.SendInteger(Connection.TRANSACTION_NONE, 4);
	    break;
    }
	    } catch (IOException e) {}
  }

  /**
   * Get this Connection's current transaction isolation mode.
   *
   * @return the current TRANSACTION_* mode value
   * @exception SQLException if a database access error occurs
   */
  public int getTransactionIsolation() throws SQLException
  {
    return vcTransactionIsolation;
  }

  /**
   * The first warning reported by calls on this Connection is
   * returned.
   *
   * <B>Note:</B> Sebsequent warnings will be changed to this
   * SQLWarning
   *
   * @return the first SQLWarning or null
   * @exception SQLException if a database access error occurs
   */
  public SQLWarning getWarnings() throws SQLException
  {
    return null;	// We handle warnings as errors
  }

  /**
   * After this call, getWarnings returns null until a new warning
   * is reported for this connection.
   *
   * @exception SQLException if a database access error occurs
   */
  public void clearWarnings() throws SQLException
  {
    // Not handles since we handle wanrings as errors
  }

  // **********************************************************
  //		END OF PUBLIC INTERFACE
  // **********************************************************

  /**
   * Send a query to the backend.  Returns one of the ResultSet
   * objects.
   *
   * <B>Note:</B> there does not seem to be any method currently
   * in existance to return the update count.
   *
   * @param sql the SQL statement to be executed
   * @return a ResultSet holding the results
   * @exception SQLException if a database error occurs
   */
  public synchronized ResultSet ExecSQL(String sql) throws SQLException
  {
    Field[] fields = null;
    Vector tuples = new Vector();
    String recv_status = null, msg;
    SQLException final_error = null;
    boolean qexit=false;
    int updateCount=1;
    int extendedCount;
    boolean extended=false;

    if (sql.length() > 1000)
      throw new SQLException("SQL Statement too long: " + sql);
    if(iddb_stream==null)
      throw new SQLException("Connection has been closed");

    try
      {
	iddb_stream.SendChar('Q');
	iddb_stream.SendString(sql);

      while (!qexit)
	{
	  int c = iddb_stream.ReceiveChar();

	  switch (c)
	    {
	    case 'T':	// MetaData Field Description
	        if (fields != null)
		        throw new SQLException("Cannot handle multiple result groups");
	        fields = ReceiveFields();
	        break;
	    case 'D':	// Data Transfer
	        if (fields == null)
		        throw new SQLException("Tuple received before MetaData");
	        tuples.addElement(iddb_stream.ReceiveTuple(fields.length, true));
	        break;
	    case 'E':	// Error Message
	        msg = iddb_stream.ReceiveString(1000);
//	        System.out.println("ada error di connection stream : "+msg);
	        final_error = new SQLException(msg);
	        qexit = true;
	        break;
	    case 'X':	// eXtended Message
	        extendedCount = iddb_stream.ReceiveInteger(4);
	        extendedData = iddb_stream.Receive(extendedCount);
	        extended=true;
	        break;
	    case 'U':	// Update Count Message
	        updateCount = iddb_stream.ReceiveInteger(4);
	        break;
	    case 'W':	// Warning Message
	        msg = iddb_stream.ReceiveString(1000);
	        final_error = new SQLException(msg);
	        qexit = true;
	        break;
	    case 'O':	// Oke Message
	        qexit = true;
	        break;
	    case 'S':	// Shutdown Message
	        qexit = true;
	        close();
	        throw new SQLException("Remote Close Connection");
	    default:
	        throw new SQLException("Unknown Response Type: " + (char)c);
	    }
	}
      } catch (IOException e) {
	throw new SQLException("I/O Error: " + e.toString());
      }
      if (final_error != null)
	throw final_error;
//	if(!extended)
      return new ResultSet(this, fields, tuples, recv_status, updateCount);
  }

  /**
   * Receive the field descriptions from the back end
   *
   * @return an array of the Field object describing the fields
   * @exception SQLException if a database error occurs
   */
  private Field[] ReceiveFields() throws SQLException
  {
    int nf=0,i;
    Field[] fields = null;
    try
    {
        nf = iddb_stream.ReceiveInteger(4);

    fields = new Field[nf];

//	System.out.println(" Jumlah field: "+nf);
    for (i = 0 ; i < nf ; ++i)
      {
	String typname = iddb_stream.ReceiveString(21);
//	System.out.print(" | "+typname);
	int typid = iddb_stream.ReceiveInteger(4);
//	System.out.print(" | "+typid);
	int typlen = iddb_stream.ReceiveInteger(4);
//	System.out.println(" | "+typlen+" | ");
	fields[i] = new Field(this, typname, typid, typlen);
      }
    } catch (IOException e)
    {
	    throw new SQLException("I/O Error: " + e.toString());
    }
//	System.out.println(" Keluar dari fields ");
    return fields;
  }

  /**
   * In SQL, a result table can be retrieved through a cursor that
   * is named.  The current row of a result can be updated or deleted
   * using a positioned update/delete statement that references the
   * cursor name.
   *
   * We support one cursor per connection.
   *
   * setCursorName sets the cursor name.
   *
   * @param cursor the cursor name
   * @exception SQLException if a database access error occurs
   */
  public void setCursorName(String cursor) throws SQLException
  {
    this.cursor = cursor;
  }

  /**
   * getCursorName gets the cursor name.
   *
   * @return the current cursor name
   * @exception SQLException if a database access error occurs
   */
  public String getCursorName() throws SQLException
  {
    return cursor;
  }

  /**
   * We are required to bring back certain information by
   * the DatabaseMetaData class.  These functions do that.
   *
   * Method getURL() brings back the URL (good job we saved it)
   *
   * @return the url
   * @exception SQLException just in case...
   */
  public String getURL() throws SQLException
  {
    return this_url;
  }

  /**
   * Method getUserName() brings back the User Name (again, we
   * saved it)
   *
   * @return the user name
   * @exception SQLException just in case...
   */
  public String getUserName() throws SQLException
  {
    return IDDB_USER;
  }

}

// ***********************************************************************

